#ifndef DOCUMENT_H
#define DOCUMENT_H

#include <string>
#include <istream>

#include "puzzle.h"
#include "board.h"
#include "actions.h"

namespace sudoku
{
  typedef Step  Hint;
  typedef StepCollection  HintCollection;
  typedef HintCollection::const_iterator  HintIterator;

  class Generator;

  class Document
  {
  public:
    typedef Board::SquareCollection  SquareCollection;
    typedef BoardCollection::size_type  size_type;
    typedef Puzzle::ColouringOption ColouringOption;
    typedef unsigned colour_type;

    typedef Puzzle::SquareCollection  RefCollection;
    typedef RefAction<Square>         SquareRefAction;

    class Moves
    {
    private:
      RefCollection  m_moves;

    public:
      void add(Square *s) { m_moves.push_back(s); }
      void removeLatest() { m_moves.pop_back(); }
      Square* getLatest() const {
          return (m_moves.empty() ? NULL : m_moves.back());
      }
      void clear() { m_moves.clear(); }
      size_type count() const { return m_moves.size(); }

      const RefCollection& getMoves() const { return m_moves; }
    };

    class Note
    {
    public:
      typedef Document::size_type  size_type;
      static const size_type MAX_MARKS = 4;

    private:
      unsigned m_mark[MAX_MARKS];
      bitmask_t m_marks;
      colour_type m_colour;

    public:
      Note();
      bool hasMark(Note::size_type i) const { return (m_mark[i] != 0); } 
      bool hasMarks() const { return (m_marks != 0); }
      bool setMark(size_type, unsigned);
      bool clearMark(size_type);

      colour_type getColour() const { return m_colour; }
      void setColour(colour_type c) { m_colour = c; }
      void clearColour() { setColour(0); }
      bool hasColour() const { return (m_colour != 0); }

      void clear();

      const unsigned operator[](int i) const { return m_mark[i]; }
      unsigned& operator[](int i) { return m_mark[i]; }
    };

    typedef std::vector<Note>  NoteCollection;

    class NoteAction
    {
    public:
      virtual ~NoteAction() {}
      virtual void operator()(Note&, Note::size_type) const = 0;
    };

    class NoteBook
    {
    public:
      typedef NoteCollection::size_type  size_type;

    private:
      const Geometry& m_geometry;
      NoteCollection m_notes;

    public:
      explicit NoteBook(const Geometry&);
      void clear();

      const Geometry& getGeometry() const { return m_geometry; }

      const Note& getNoteFor(const Square &s) const {
          return m_notes[s.getSquareIndex()];
      }
      Note& getNoteFor(const Square &s) {
          return m_notes[s.getSquareIndex()];
      }

      bool hasMark(const Square&, unsigned) const;
      bool setMark(const Square&, size_type, unsigned);
      bool clearMark(const Square&, size_type);

      colour_type getColour(const Square &s) const;
      void setColour(const Square&, colour_type);
      void clearColour(const Square &s) { setColour(s, 0); }

      size_type size() const { return m_notes.size(); }

      const Note& operator[](int i) const { return m_notes[i]; }
      Note& operator[](int i) { return m_notes[i]; }

      void do_for_each(const NoteAction&);

      bool readMarks(std::istream&);
      bool readColour(std::istream&);
    };

  private:
    BoardCollection m_solutions;
    Board*          m_original;
    Board*          m_board;

    bool            m_hasLogical;
    StepCollection  m_solutionLog;

    Moves           m_moves;
    RefCollection   m_givenHints;

    HintCollection  m_allHints;
    HintIterator    m_hint;

    bool            m_hintsAreStale;

    NoteBook*       m_notebook;

    unsigned        m_maxHiddenSubsetSize;
    unsigned        m_maxNakedSubsetSize;
    unsigned        m_maxPatternSize;
    ColouringOption m_colouring;
    unsigned        m_wings;

    bool hintsAreStale() const { return m_hintsAreStale; }
    void staleHints() { m_hintsAreStale = true; }
    void freshHints() { m_hintsAreStale = false; }

    bool solve(Board&, BoardCollection&, StepCollection* =NULL);
    bool solve();

    void loadPlayState(std::istream &istr);
    void generateHints();

    static bool deleteMistakes(HintCollection&,
                               HintCollection::iterator,
                               HintCollection::iterator&);
    static void deleteMistakes(HintCollection&);

    static void throw_illegal_option(const char*, const Square&);

  public:
    static const size_type UPPER_SOLUTION_LIMIT = 5;

    Document();
    ~Document();

    Board* getBoard() const { return m_board; }
    Board* getOriginalBoard() const { return m_original; }
    const Geometry& getGeometry() const { return m_board->getGeometry(); }
    NoteBook* getNoteBook() const { return m_notebook; }

    const Note& getNoteFor(const Square &s) const {
        return m_notebook->getNoteFor(s);
    }
    Note& getNoteFor(const Square &s) {
        return m_notebook->getNoteFor(s);
    }

    const BoardCollection& getSolutions() const { return m_solutions; }
    size_type getSolutionCount() const { return m_solutions.size(); }
    bool hasLogicalSolution() const { return m_hasLogical; }

    bool hasUniqueSolution() const { return (m_solutions.size() == 1); }
    const Board& getUniqueSolution() const { return m_solutions.front(); }

    const StepCollection& getSolutionLog() const { return m_solutionLog; }

    const char* getRating() const { return getRating(m_solutions); }
    const std::string& getStatus(std::string &s) const {
        return getStatus(s, hasLogicalSolution(), getSolutionCount());
    }

    bool isComplete() const {
        return isComplete( getBoard() );
    }

    unsigned getSubsetSizeLimit() const {
        return getGeometry().getSubsetSizeLimit();
    }
    unsigned getMaxHiddenSubsetSize() const { return m_maxHiddenSubsetSize; }
    void setMaxHiddenSubsetSize(unsigned s) { m_maxHiddenSubsetSize = s; }

    unsigned getMaxNakedSubsetSize() const { return m_maxNakedSubsetSize; }
    void setMaxNakedSubsetSize(unsigned s) { m_maxNakedSubsetSize = s; }

    unsigned getPatternSizeLimit() const {
        return getGeometry().getPatternSizeLimit();
    }
    unsigned getMaxPatternSize() const { return m_maxPatternSize; }
    void setMaxPatternSize(unsigned s) { m_maxPatternSize = s; }

    ColouringOption getColouring() const { return m_colouring; }
    void setColouring(ColouringOption c) { m_colouring = c; }

    unsigned getWings() const { return m_wings; }
    void setWings(unsigned w) { m_wings = w; }

    Board& recreate(const Geometry&);

    void load(std::istream&);
    void savePlay(std::ostream&) const;
    void savePuzzle(std::ostream&) const;
    bool resetSolution();
    void reset();
    void finalise();
    void generate(int target=0);

    void addMove(Square&, unsigned);
    Square* removeLatestMove();
    size_type countMoves() const { return m_moves.count(); }

    void do_for_each_move(const SquareRefAction&) const;
    void do_for_each_hint(const SquareRefAction&) const;
    void do_for_each_note(const NoteAction &action) {
        return m_notebook->do_for_each(action);
    }

    const Hint *getHint();
    bool hasHints() const { return (m_hint != m_allHints.end()); }
    void addHint(Square&, unsigned);

    bool addMark(const Square&, Note::size_type, unsigned);
    void clearMark(const Square&, Note::size_type);
    void clearMarks(const Square&);
    bool hasMark(const Square&, Note::size_type) const;
    colour_type getColour(const Square&) const;

    bool validate(const Board*, BoardCollection&);
    bool isErrorMove(const Square&) const;
    static bool isErrorMove(const Board&, const Square&);

    std::string& generateFileName(std::string&) const;
    static const char* getRating(const BoardCollection&);
    static void validate(const Board*);
    static const std::string& getStatus(std::string&, bool, size_type);
    static bool isComplete(const Board*);
  };
}

#endif

